#include <LPC11xx.h>                                   // LPC1100 definitions

#include "global.h"


const WORD fade_time[16] = {0,70,100,141,200,283,400,566,800,1131,1600,2263,3200,4526,6400,9051};
const BYTE fade_rate[16] = {0,72,51,36,25,18,13,9,6,4,3,2,2,1,1,1};

static BYTE fade_stop;                                 // level after fading
static BYTE fade_start;
static WORD fade_total;
static WORD fade_timer;

BYTE state = ST_INIT;                                  // state of the fade timer
BYTE level = 0;                                        // current light level
BYTE limit_error = 0;                                  // 1 if the last requested level was out of limit


void SetLevel(BYTE new_level)
{
    state = ST_IDLE;
    limit_error = 0;
    if (new_level > 0 && new_level < MIN_LEVEL)
    {
        new_level = MIN_LEVEL;
        limit_error = 1;
    }
    if (new_level > MAX_LEVEL && new_level < 255)
    {
        new_level = MAX_LEVEL;
        limit_error = 1;
    }
    if (new_level != 255)                              // mask
    {
        LPC_TMR32B1->MR0 = (~new_level) & 0x00FF;      // set PWM duty cycle
        level = new_level;
    }
}

void FadeTime(BYTE new_level)                          // Dim (up/down) to new level using the FADE TIME
{
    if (new_level != 255)                              // mask
    {
        if (FADE_TIME)
        {                                              // Fade with fade time
            fade_stop  = new_level;
            fade_start = level;
            fade_total = fade_time[FADE_TIME];
            fade_timer = fade_total;
            if (level > new_level)
                state = FADE_DOWN;
            else
                state = FADE_UP;
        }
        else
            SetLevel(new_level);                       // No fade
    }
}

void FadeUp(void)                                      // Dim up 200 msec using the FATE RATE
{
    if (level != 0 && level != MAX_LEVEL)              // no change if lamp off or at MAX_LEVEL
    {
        fade_start = level;
        if (fade_rate[FADE_RATE] > (MAX_LEVEL - level))
        {
            fade_stop  = MAX_LEVEL;
            fade_total = (MAX_LEVEL - level)*20/fade_rate[FADE_RATE];
        }
        else
        {
            fade_stop  = level + fade_rate[FADE_RATE];
            fade_total = 20;                           // 200 ms
        }
        fade_timer = fade_total;
        state = FADE_UP;
    }    
}

void FadeDown(void)                                    // Dim down 200 msec using the FATE RATE
{
    if (level != 0 && level != MIN_LEVEL)              // no change if lamp off or at MIN_LEVEL
    {
        fade_start = level;
        if (fade_rate[FADE_RATE] > (level - MIN_LEVEL))
        {
            fade_stop  = MIN_LEVEL;
            fade_total = (level - MIN_LEVEL)*20/fade_rate[FADE_RATE];
        }
        else
        {
            fade_stop  = level - fade_rate[FADE_RATE];
            fade_total = 20;                           // 200 ms
        }
        fade_timer = fade_total;
        state = FADE_DOWN;
    }    
}

BYTE LampFailure(void)                                 // Test lamp, return 1 if bad lamp
{
    return 0;                                          // lamp is okay
}

void SysTick_Handler (void)                            // SysTick Timer ISR every 10 msec
{
  LONG diff;

    if (rep_timeout > 0) rep_timeout --;

    if (add_timeout > 0) add_timeout --;

    if (fade_timer  > 0) fade_timer --;

    switch (state)
    {
      case ST_INIT:   if (fade_timer == 0)
                          SetLevel(POWER_ON_LEVEL);
                      break;
      case FADE_UP:   diff = ((fade_stop - fade_start)*(fade_total - fade_timer))/fade_total;
                      SetLevel(fade_start + diff);
                      if (fade_timer)
                          state = FADE_UP;
                      break;
      case FADE_DOWN: diff = ((fade_start - fade_stop)*(fade_total - fade_timer))/fade_total;
                      SetLevel(fade_start - diff);
                      if (fade_timer)
                          state = FADE_DOWN;
      default:        break;
    }
}

void Ballast_Init(void)
{
    fade_timer = 60;                                   // 600 msec for power-up

    LPC_SYSCON->SYSTICKCLKDIV = 1;                     // Set systick clock = sys clock
    SysTick_Config(SystemFrequency/100);               // Generate interrupt each 10 ms

    LPC_SYSCON->SYSAHBCLKCTRL |= (1<<10);              // CT32B1 clock enable
    LPC_IOCON->JTAG_TDO_PIO1_1 = 0xC3;                 // select P1.1 as CT32B1_MAT0 output

    LPC_TMR32B1->MR3  = 255;                           // -> PMW base frequency = SystemFrequency / 255
    LPC_TMR32B1->MR0  = 255;                           // light off
    LPC_TMR32B1->MCR  = 0x0400;                        // reset TC on MR3
    LPC_TMR32B1->PWMC = 0x00000009;                    // enable PWM output (P1.1)
    LPC_TMR32B1->TCR  = 0x01;                          // start timer 2
}
