#include <LPC11xx.h>                               // LPC1100 definitions

#include "global.h"
#include "command.h"

BYTE rep_timeout = 0;                              // used for 100 ms repeat message timeout (see ballast.c)
LONG add_timeout = 0;                              // used for 15 min addressing commands timeout (see ballast.c)

static BYTE dtr;                                   // Data Transfer Register
static BYTE repeat_address;                        // address for repeation
static BYTE repeat_command;                        // command for repeation
static BYTE power_failure  = 1;                    // true until certain commands are received
static LONG search_address = 0xFFFFFF;


static void HandleSpecialCommands(void)                            // Handle special commands
{
  static BYTE withdraw;
  static BYTE selection;
  static LONG random;

    if (address >= 0xA7 && address <= 0xBD && add_timeout == 0)
    {
        return;                                                    // Not valid until INITIALISE has been received
    }   

    switch (address)
    {
      case TERMINATE:      add_timeout = 0;                        // terminate special mode processes
                           break;
      case DATA_TRANS_REG: dtr = command;                          // store 8 bit value to DTR
                           break;
      case INITIALISE:     if (command == 0 || command == ((SHORT_ADDRESS<<1)|1))
                           {
                               add_timeout = 90000;                // accept addressing commands for the next 15 minutes
                               withdraw    = 0;
                               selection   = 0;
                           }
                           break;
      case RANDOMISE:      FlashRandom(LPC_TMR32B1->TC);           // generate new random address
                           break;
      case COMPARE:        random = RANDOM_ADDRESS_H<<16 | RANDOM_ADDRESS_M<<8 | RANDOM_ADDRESS_L;
                           if ((random < search_address) || ((random == search_address) && !withdraw))
                           {
                               answer = 0xFF;                      // yes
                               f_dalitx = 1;
                           }
                           break;
      case WITHDRAW:       withdraw = 1;                           // exclude ballast from compare process
                           break;
      case SEARCHADDRH:    search_address &= 0x00FFFF;             // bit 16-23 of the search address
                           search_address |= command<<16;
                           break;
      case SEARCHADDRM:    search_address &= 0xFF00FF;             // bit 8-15 of the search address
                           search_address |= command<<8;
                           break;
      case SEARCHADDRL:    search_address &= 0xFFFF00;             // bit 0-7 of the search address
                           search_address |= command;
                           break;
      case PROGRAM_SHORT:  random = RANDOM_ADDRESS_H<<16 | RANDOM_ADDRESS_M<<8 | RANDOM_ADDRESS_L;
                           if ((selection && LampFailure()) || random == search_address)
                           {
                               if (command == 255)
                               {
                                   FlashWrite(9, 255);
                               }
                               else if (((command & 0x7E) >> 1) < 64)
                               {
                                   FlashWrite(9, (command & 0x7E) >> 1);
                               }
                           }
                           break;
      case VERIFY_SHORT:   if (SHORT_ADDRESS == ((command & 0x7E) >> 1))
                           {
                               answer = 0xFF;                      // yes, if short address is equal
                               f_dalitx = 1;
                           }
                           break;
      case QUERY_SHORT:    random = RANDOM_ADDRESS_H<<16 | RANDOM_ADDRESS_M<<8 | RANDOM_ADDRESS_L;
                           if ((selection && LampFailure()) || random == search_address)
                           {
                               answer = SHORT_ADDRESS;             // send the short address
                               f_dalitx = 1;
                           }
                           break;
      case PHYSICAL_SEL:   selection = 1;                          // set physical selection mode
                           break;
      case ENABLE_TYPE:    break;                                  // not implemented
      default:             break;
    }
}

static void HandleNormalCommands(void)                             // Handle normal commands
{
    switch (command)
    {
      case OFF:             power_failure = 0;                     // lamp off immediately, no fading
                            SetLevel(0);
                            break;
      case UP:              FadeUp();                              // Dim up 200 ms, using the fade rate
                            break;
      case DOWN:            FadeDown();                            // Dim down 200 ms, using the fade rate
                            break;
      case STEP_UP:         if (level != 0 && level < MAX_LEVEL)   // power level one step higher, no fade
                                SetLevel(level + 1);
                            break;
      case STEP_DOWN:       if (level > MIN_LEVEL)                 // power level one step lower, no fade
                                SetLevel(level - 1);
                            break;
      case RECALL_MAX:      power_failure = 0;                     // power level to max level, no fade
                            SetLevel(MAX_LEVEL);
                            break;
      case RECALL_MIN:      power_failure = 0;                     // power level to min level, no fade
                            SetLevel(MIN_LEVEL);
                            break;
      case DOWN_AND_OFF:    power_failure = 0;                     // one step lower, and off if at min level
                            if (level == MIN_LEVEL)
                                SetLevel(0);
                            else if (level != 0)
                                SetLevel(level - 1);
                            break;
      case ON_AND_UP:       power_failure = 0;                     // one step higher / if off to min level
                            if (level == 0)
                                SetLevel(MIN_LEVEL);
                            else if (level != MAX_LEVEL)
                                SetLevel(level + 1);
                            break;
      case RESET:           power_failure = 0;                     // restore default values to flash memory
                            search_address = 0xFFFFFF;
                            FlashReset();
                            SetLevel(254);
                            break;
      case STORE_ACTUAL:    dtr = level;                           // store actual level in the dtr
                            break;
      case STORE_MAX:       if (dtr > MIN_LEVEL && dtr < 255)      // store the dtr as max level
                                FlashWrite(3, dtr);
                            break;
      case STORE_MIN:       if (dtr > 0 && dtr < MAX_LEVEL)        // store the dtr as min level
                                FlashWrite(2, dtr);
                            break;
      case STORE_FAILURE:   FlashWrite(1, dtr);                    // store the dtr as system failure level
                            break;
      case STORE_POWER_ON:  if (dtr > 0 && dtr < 255)              // store the dtr as power on level
                                FlashWrite(0, dtr);
                            break;
      case STORE_FADE_TIME: FlashWrite(5, dtr & 0x0F);             // store the dtr as fade time
                            break;
      case STORE_FADE_RATE: FlashWrite(4, dtr & 0x0F);             // store the dtr as fade rate
                            break;
      case STORE_SHORT_ADD: if (dtr == 255)                        // store the dtr as short address
                                FlashWrite(9, 255);                // 'mask'
                            else if (((dtr & 0x7E)>>1) < 64)
                                FlashWrite(9, (dtr & 0x7E) >> 1);
                            break;
      case QUERY_STATUS:    answer = 0;                            // answer is the status of the ballast
                            if (LampFailure())
                                answer |= 0x02;
                            if (level)
                                answer |= 0x04;
                            if (limit_error)
                                answer |= 0x08;
                            if (state == FADE_UP || state == FADE_DOWN)
                                answer |= 0x10;
                            if (state == ST_INIT)
                                answer |= 0x20;
                            if (SHORT_ADDRESS == 255)
                                answer |= 0x40;
                            if (power_failure)
                                answer |= 0x80;
                            f_dalitx = 1;
                            break;
      case QUERY_BALLAST:   answer = 0xFF;                         // ask is ballast is available (at given address)
                            f_dalitx = 1;
                            break;
      case QUERY_LFAILURE:  break;                                 // ask if there's a lamp failure, not implemented
      case QUERY_POWER_ON:  if (level)                             // ask if lamp is operating at given address
                            {
                                answer = 0xFF;                     // yes
                                f_dalitx = 1;
                            }
                            break;
      case QUERY_LIMIT_ERR: if (limit_error)                       // ask if last requested power level could not
                            {                                      // be met (above max or below min level)
                                answer = 0xFF;                     // yes
                                f_dalitx = 1;
                            }
                            break;
      case QUERY_RESET:     if (state == ST_INIT)                  // ask if ballast is in reset state
                            {
                                answer = 0xFF;                     // yes
                                f_dalitx = 1;
                            }
                            break;
      case QUERY_MISSING:   if (SHORT_ADDRESS == 255)              // ask if ballast has no short address
                            {
                                answer = 0xFF;                     // yes
                                f_dalitx = 1;
                            }
                            break;
      case QUERY_VERSION:   answer = VERSION_NUMBER;               // ask for the version number
                            f_dalitx = 1;
                            break;
      case QUERY_CONTENT:   answer = dtr;                          // ask for the content of the dtr
                            f_dalitx = 1;
                            break;
      case QUERY_DEVICE:    answer = 0;                            // ask for device type, 0 = standard device
                            f_dalitx = 1;
                            break;
      case QUERY_MINIMUM:   answer = PHY_MIN_LEVEL;                // ask for the physically lowest level
                            f_dalitx = 1;
                            break;
      case QUERY_PFAILURE:  if (power_failure)                     // ask for the power failure status
                            {
                                answer = 0xFF;                     // yes
                                f_dalitx = 1;
                            }
                            break;
      case QUERY_ACTUAL:    answer = level;                        // ask for the actual power level
                            f_dalitx = 1;
                            break;
      case QUERY_MAX_LEVEL: answer = MAX_LEVEL;                    // ask for the maximum level
                            f_dalitx = 1;
                            break;
      case QUERY_MIN_LEVEL: answer = MIN_LEVEL;                    // ask for the minimum level
                            f_dalitx = 1;
                            break;
      case QUERY_POWER_L:   answer = POWER_ON_LEVEL;               // ask for the power on level
                            f_dalitx = 1;
                            break;
      case QUERY_FAILURE_L: answer = SYS_FAILURE_LEVEL;            // ask for the system failure level
                            f_dalitx = 1;
                            break;
      case QUERY_FADE:      answer = FADE_TIME << 4 | FADE_RATE;   // ask for the fade time and fade rate
                            f_dalitx = 1;
                            break;
      case QUERY_GR_0_7:    answer = GROUP_0_7;                    // to what group (0 to 7) does ballast belong
                            f_dalitx = 1;
                            break;
      case QUERY_GR_8_15:   answer = GROUP_8_15;                   // to what group (8 to 15) does ballast belong
                            f_dalitx = 1;
                            break;
      case QUERY_ADDRESS_H: answer = RANDOM_ADDRESS_H;             // ask for 8 high bits of the random address
                            f_dalitx = 1;
                            break;
      case QUERY_ADDRESS_M: answer = RANDOM_ADDRESS_M;             // ask for 8 mid bits of the random address
                            f_dalitx = 1;
                            break;
      case QUERY_ADDRESS_L: answer = RANDOM_ADDRESS_L;             // ask for 8 low bits of the random address
                            f_dalitx = 1;
      default:              break;
    }

    switch (command & 0xF0)
    {
      case GO_TO_SCENE:     power_failure = 0;                     // set power level for scene, using fade time
                            FadeTime(*(SCENE + (command & 0x0F)));
                            break;
      case STORE_SCENE:     FlashWrite(12+(command & 0x0F), dtr);  // store dtr as new level for scene xxxx
                            break;
      case REMOVE_SCENE:    FlashWrite(12+(command & 0x0F), 255);  // remove the ballast from scene xxxx
                            break;
      case ADD_TO_GROUP:    if (command < 0x68)                    // add the ballast to group xxxx
                                FlashWrite(10, GROUP_0_7 | 1 << (command & 0x0F));
                            else
                                FlashWrite(11, GROUP_8_15 | 1 << ((command & 0x0F)-8));
                            break;
      case REMOVE_GROUP:    if (command < 0x78)                    // remove the ballast from group xxxx
                                FlashWrite(10, GROUP_0_7 & ~(1<<(command & 0x0F)));
                            else
                                FlashWrite(11, GROUP_8_15 & ~(1<<((command & 0x0F)-8)));
                            break;
      case QUERY_SCENE:     answer = *(SCENE + (command & 0x0F));  // ask for power level of scene xxxx
                            f_dalitx = 1;
      default:              break;
    }
}

void DALI_ProcesCommand(void)
{
  WORD group;

    if ((address & 0xE1)!= 0xA1 &&                                 //     not special command 101CCCC1 ?
	    (address & 0xE1)!= 0xC1 &&                                 // and not special command 100CCCC1 ?
		(address & 0xFE)!= 0xFE)                                   // and not broadcast message ?
    {
        if ((address & 0xE0) == 0x80)                              // group address 100AAAAS ?
        {
            group = GROUP_8_15 << 8 | GROUP_0_7;
            if ((group >> ((address & 0x1E) >> 1) & 1) == 0)       // right group address ?
            {
                return;                                            // no, not in this group
            }
        }
        else if ((address & 0x80) == 0)                            // short address 0AAAAAAS ?
        {
            if (((address & 0x7E)>>1)!= SHORT_ADDRESS)
            {   
                return;                                            // no, wrong short address
            }
        }
        else
            return;                                                // invalid address
    }
    
    if (rep_timeout != 0)                                          // check if repeated message
    {
        if (address != repeat_address || command != repeat_command)
        {
            rep_timeout = 0;                                       // ignore message and clear repeat flag
            return;
        }
    }

    if ((address & 0x01) == 0)                                     // direct arc power control command
    {
        power_failure = 0;
        FadeTime(command);
        return;
    }

    if ((address & 0xE0) == 0xA0 || (address & 0xE0) == 0xC0)      // special command ?
    {
        if (address == INITIALISE || address == RANDOMISE)
        {
            if (rep_timeout == 0)                                  // these commands shall be repeated
            {
                repeat_address = address;
                repeat_command = command;
                rep_timeout = 10;                                  // 10* 10 ms = 100 ms
                return;
            }
        }
    }
    else                                                           // normal command
    {
        if (command >= 32 && command <= 128)                       // configuration command ?
        {
            if (rep_timeout == 0)                                  // yes, they shall be repeated
            {
                repeat_address = address;
                repeat_command = command;
                rep_timeout = 10;                                  // 10* 10 ms = 100 ms
                return;
            }
        }
    }
    rep_timeout = 0;

    if ((address & 0xE0) == 0xA0 || (address & 0xE0) == 0xC0)
        HandleSpecialCommands();                                   // handle special commands
    else
        HandleNormalCommands();                                    // handle normal commands
}
