/********************************************************************
;
;  DALI slave
;
;  For receive, this module uses CT32B0-CAP0 input (capture and interrupt on both edges)
;  CAP0 (P1.5) is connected to P2.6 (to check high / low level by software)
;  DALI forward frame format (INVERTED input, because of opto-coupler):
;
;      | S |        8 address bits         |        8 command bits         | stop  |
;      | 1 | 1 | 0 | 0 | 0 | 0 | 0 | 1 | 1 | 0 | 1 | 1 | 1 | 1 | 0 | 0 | 0 |   |   |
;
;      +-+ +-+   +-+ +-+ +-+ +-+ +---+ +-+   +---+ +-+ +-+ +-+   +-+ +-+ +-+
;      | | | |   | | | | | | | | |   | | |   |   | | | | | | |   | | | | | |
;   ---+ +-+ +---+ +-+ +-+ +-+ +-+   +-+ +---+   +-+ +-+ +-+ +---+ +-+ +-+ +------------
;
;      |2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|  4TE  |
;
;
;  For transmit, this module uses GPIO - P2.0 (DALI send pin)
;  DALI backward frame format:
;
;                    | S |         8 data bits           | stop  |
;                    | 1 | 1 | 0 | 0 | 0 | 0 | 0 | 1 | 1 |   |   |
;
;    +---------------+ +-+ +---+ +-+ +-+ +-+ +-+   +-+ +-------------
;    |               | | | |   | | | | | | | | |   | | |
;   -+               +-+ +-+   +-+ +-+ +-+ +-+ +---+ +-+
;
;    |4 + 7 to 22 TE |2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|2TE|  4TE  |
;
;  2TE = 834 usec (1200 bps)
;
*******************************************************************************************/

#include <LPC11xx.h>                               // LPC1100 definitions

#include "global.h"

#define TE             834/2                       // half bit time = 417 usec
#define MIN_TE         TE - TE/10                  // minimum half bit time
#define MAX_TE         TE + TE/10                  // maximum half bit time
#define MIN_2TE        2*TE - TE/10                // minimum full bit time
#define MAX_2TE        2*TE + TE/10                // maximum full bit time

static int  low_time;                              // captured puls low time
static int  high_time;                             // captured puls high time

static BYTE value;                                 // used for dali send bit
static BYTE position;                              // keeps track of sending bit position
static BYTE previous;                              // previous received bit
static LONG frame;                                 // holds the received dali forward frame


static void DALI_Shift_Bit(BYTE val)
{
    if (frame & 0x10000)                           // frame full ?
        frame = 0;                                 // yes, ERROR
    else
        frame = (frame << 1) | val;                // shift bit
}

/************************************************************************
; DALI_Decode (we only take action at a falling edge)
;
; previous bit     Low Time       High Time      Action     New previous bit
; -------------------------------------------------------------------
;     0               0               0          Shift 0          0
;     0               0               1          Shift 0,1        1
;     0               1               0          -ERROR-          *
;     0               1               1          -ERROR-          *
;     1               0               0          Shift 1          1
;     1               0               1          -ERROR-          *
;     1               1               0          Shift 0          0
;     1               1               1          Shift 0,1        1
;
***********************************************************************/
static void DALI_Decode(void)
{
  BYTE action;

    action = previous << 2;

    if ((high_time > MIN_2TE) && (high_time < MAX_2TE))
        action = action | 1;                       // high_time = long
    else if (!((high_time > MIN_TE) && (high_time < MAX_TE)))
    {
        frame = 0;                                 // DALI ERROR
        return;
    }

    if ((low_time > MIN_2TE) && (low_time < MAX_2TE))
        action = action | 2;                       // low_time = long
    else if (!((low_time > MIN_TE) && (low_time < MAX_TE)))
    {
        frame = 0;                                 // DALI ERROR
        return;
    }

    switch (action)
    {
      case 0:  DALI_Shift_Bit(0);                  // short low, short high, shift 0
               break;
      case 1:  DALI_Shift_Bit(0);                  // long low, short high, shift 0,1
               DALI_Shift_Bit(1);
               previous = 1;                       // new half bit is 1
               break;
      case 2:                                      // short low, long high, ERROR
      case 3:  frame = 0;                          // long low, long high, ERROR
               break;
      case 4:  DALI_Shift_Bit(1);                  // short low, short high, shift 1
               break;
      case 5:  frame = 0;                          // long low, short high, ERROR
               break;
      case 6:  DALI_Shift_Bit(0);                  // short low, long high, shift 0
               previous = 0;                       // new half bit is 0
               break;
      case 7:  DALI_Shift_Bit(0);                  // long low, long high, shift 0,1
               DALI_Shift_Bit(1);
      default: break;                              // invalid
    }
}

void TIMER32_0_IRQHandler(void)
{
    LPC_TMR32B0->TC = 0;                           // reset timer
												   
    if (LPC_TMR32B0->IR & 1)                       // 11TE timeout for settling time ?
    {                                              // minimum time after last DALI pulse
        if (frame & 0x10000)                       // frame completed ?
        {
            command = (BYTE)frame;                 // OK ! save command byte
            address = frame >> 8;                  //      save address byte
            f_dalirx = 1;                          // set flag to signal application
            LPC_TMR32B0->MCR = 0x0000;             // disable interrupt on MR0
        }
        frame = 0;
        LPC_TMR32B0->IR = 0x01;                    // clear MR0 interrupt flag
    }
    else if (LPC_TMR32B0->IR & 2)                  // match 2 interrupt for DALI send
    {
        if (value)
            LPC_GPIO2->DATA |= (1<<0);             // P2.0 DALI output pin high
        else
            LPC_GPIO2->DATA &= ~(1<<0);            // P2.0 DALI output pin low

        if (position == 0)                         // 0TE second half of start bit
        {
            value = 1;                             // second half of start bit = 1
        }
        else if (position < 17)                    // 4 + (1TE to 16TE), so data bits
        {
            value = (answer >> ((16 - position)/2)) & 1;
            if (position & 1)                      // first or second half of data bit
            {
                value = !value;                    // invert if first half
            }
        }
        else if (position == 17)                   // 4 + 17TE start of stop bit (4TE)
        {                                          // and start of minimum settling time (7TE)
            value = 1;
        }
        else if (position == 22)                   // finished with DALI Send
        {
            LPC_TMR32B0->MCR = 0x0000;             // disable interrupt on MR1
            LPC_TMR32B0->CCR = 0x0007;             // capture on both edges and enable the interrupt
        }
        position++;
        LPC_TMR32B0->IR = 0x02;                    // reset interrupt flag
    }
    else                                           // capture interrupt for DALI receive
    {
        LPC_TMR32B0->MCR = 0x0003;                 // interrupt on MR0, reset timer on match 0
        if (LPC_GPIO2->MASKED_ACCESS[(1<<6)])      // P2.6 (= P1.5) to check rising or falling edge
            low_time = LPC_TMR32B0->CR0;           // rising, so capture low time
        else
        {
            if (frame != 0)                        // not first pulse ?
            {
                high_time = LPC_TMR32B0->CR0;      // falling, so capture high time
                DALI_Decode();                     // decode received bit
            }
            else
            {
                previous = 1;                      // first pulse, so shift 1
                DALI_Shift_Bit(1);
            }
        }
        LPC_TMR32B0->IR = 0x10;                    // reset interrupt flag
    }
}

void DALI_SendAnswer(void)
{
    LPC_TMR32B0->CCR = 0x0000;                     // disable capture interrupt
    LPC_TMR32B0->MCR = 0x0018;                     // interrupt on MR1, reset timer on match 1
    LPC_TMR32B0->TC  = 0;                          // reset timer
    value    = 0;                                  // first half half start bit low
    position = 0;  
}

void DALI_Init(void)
{
    LPC_SYSCON->SYSAHBCLKCTRL |= (1<<9);           // CT32B0 clock enable

    NVIC_EnableIRQ(TIMER_32_0_IRQn);               // enable Timer 0 interrupt

    LPC_GPIO2->DIR  |= (1<<0);                     // P2.0 = DALI send pin
    LPC_GPIO2->DATA |= (1<<0);                     // P2.0 high

    LPC_IOCON->PIO1_5 = 0x52;                      // P1.5 as CT32B0.CAP0 = DALI receive pin

    LPC_TMR32B0->PR = (SystemFrequency/1000000)-1; // prescaler: timer runs at 1 MHz

    LPC_TMR32B0->MR0 = 11*TE;                      // (4+7)TE high (idle) Timeout
    LPC_TMR32B0->MR1 = TE;                         // half bit time ~ 2400 Hz
    LPC_TMR32B0->CCR = 0x0007;                     // capture on both edges and enable the interrupt
    LPC_TMR32B0->TC  = 0;                          // reset timer
    LPC_TMR32B0->TCR = 1;                          // start timer
}
